﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;

using Moq;

using TicTacToe.Model.Interfaces;
using TicTacToe.Model.Model;
using TicTacToe.UnitTest.Helpers;

namespace TicTacToe.Model.UnitTest.ModelTests
{
    [TestClass]
    public class GameTest
    {
        #region Ctor tests

        [TestMethod, TestCategory("Game")]
        public void CtorAllArgumentsPassedInstanceInitiated()
        {
            var target = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());

            Assert.IsNotNull(target);
        }

        [TestMethod, TestCategory("Game")]
        public void CtorPlayer1Initiated()
        {
            var p1 = Guid.NewGuid();
            var target = TestTargetInitializer.BuildGame(p1, Guid.NewGuid(), Guid.NewGuid());

            Assert.AreEqual(p1, target.Player1Id);
        }

        [TestMethod, TestCategory("Game")]
        public void CtorPlayer2Initiated()
        {
            var p2 = Guid.NewGuid();
            var target = TestTargetInitializer.BuildGame(Guid.NewGuid(), p2, Guid.NewGuid());

            Assert.AreEqual(p2, target.Player2Id);
        }

        [TestMethod, TestCategory("Game")]
        public void CtorGameIdInitiated()
        {
            var gameId = Guid.NewGuid();
            var target = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), gameId);

            Assert.AreEqual(gameId, target.GameId);
        }

        [TestMethod, TestCategory("Game")]
        public void CtorNextPlayerIdSetToPlayer1()
        {
            var p1 = Guid.NewGuid();
            var target = TestTargetInitializer.BuildGame(p1, Guid.NewGuid(), Guid.NewGuid());

            Assert.AreEqual(p1, target.NextStepPlayerId);
        }

        [TestMethod, TestCategory("Game")]
        public void CtorFieldInitialized()
        {
            var target = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());

            Assert.IsNotNull(target.Field);
        }

        [TestMethod, TestCategory("Game")]
        public void CtorGameStatusSetToPlayerOneTurn()
        {
            var target = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());

            Assert.AreEqual(GameStatus.PlayerOneTurn, target.GameStatus);
        }

        [TestMethod, TestCategory("Game")]
        public void CtorCreatedDateTimeIsInitialized()
        {
            var target = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());

            Assert.AreNotEqual(DateTime.MinValue, target.CreatedDateTime);
        }

        [TestMethod, TestCategory("Game")]
        [ExpectedException(typeof(ArgumentException))]
        public void CtorPlayer1IdSetToEmptyThrowException()
        {
            var target = TestTargetInitializer.BuildGame(Guid.Empty, Guid.NewGuid(), Guid.NewGuid());
            
            Assert.Fail("Exception is expected");
        }

        [TestMethod, TestCategory("Game")]
        [ExpectedException(typeof(ArgumentException))]
        public void CtorPlayer2IdSetToEmptyThrowException()
        {
            var target = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.Empty, Guid.NewGuid());

            Assert.Fail("Exception is expected");
        }

        [TestMethod, TestCategory("Game")]
        [ExpectedException(typeof(ArgumentException))]
        public void CtorGameIdSetToEmptyThrowException()
        {
            var target = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.Empty, Guid.NewGuid());

            Assert.Fail("Exception is expected");
        }


        #endregion 
        
        #region NextStepOuterField tests

        [TestMethod]
        public void NextStepOuterFieldMakeStepAndDestinationFieldHasFreeSpaceNotNull()
        {
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());

            game.MakeStep(game.Player1Id, 0, 2);

            Assert.AreEqual(2, game.NextStepOuterField);
        }

        [TestMethod]
        public void NextStepOuterFieldMakeStepDestinationFieldHasNoFreeSpaceNull()
        {
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            game.Field[2].Fill(@"oxx
                                 oxo
                                 xoo");
            game.MakeStep(game.Player1Id, 0, 2);

            Assert.IsNull(game.NextStepOuterField);
        }

        #endregion

        #region GameStatus tests

        [TestMethod]
        public void GameStatusXWinTheGameStatusSetToFirstPlayerWin()
        {
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            game.Field.Fill(@"oxx
                              oxo
                              -oo");

            game.Field[6].Fill(@"oxx
                                 oxo
                                 -oo");

            game.MakeStep(game.Player1Id, 6, 6);

            Assert.AreEqual(GameStatus.PlayerOneWin, game.GameStatus);
        }

        [TestMethod]
        public void GameStatusOWinTheGameStatusSetToFirstPlayerWin()
        {
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            game.Field.Fill(@"oxx
                              oxo
                              -oo");

            game.Field[6].Fill(@"oxx
                                 oxo
                                 -oo");

            game.MakeStep(game.Player2Id, 6, 6);

            Assert.AreEqual(GameStatus.PlayerTwoWin, game.GameStatus);
        }

        [TestMethod]
        public void GameStatusGameInitiatedSetToFirstPlayerTurnn()
        {
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            
            Assert.AreEqual(GameStatus.PlayerOneTurn, game.GameStatus);
        }

        [TestMethod]
        public void GameStatusXMakeStepSetToPlayerTwoTurn()
        {
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
           
            game.MakeStep(game.Player1Id, 6, 6);

            Assert.AreEqual(GameStatus.PlayerTwoTurn, game.GameStatus);
        }

        [TestMethod]
        public void GameStatusXAndOMakeStepSetToPlayerOneTurn()
        {
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());

            game.MakeStep(game.Player1Id, 6, 6);
            game.MakeStep(game.Player2Id, 6, 1);

            Assert.AreEqual(GameStatus.PlayerOneTurn, game.GameStatus);
        }

        #endregion

        #region MakeStep tests

        [TestMethod]
        public void MakeStepPlayerMakeStepReturnDrawGameStatusStatus()
        {
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            game.Field.Fill(@"-ox
                              xoo 
                              ox-");
            game.Field[0].Fill(@"-oo
                                 xxo 
                                 xx-");

            var actual = game.MakeStep(game.Player1Id, 0, 0);

            Assert.AreEqual(GameStepStatus.Draw, actual);
        }

        #endregion

        #region IsValidPlayer tests

        [TestMethod]
        public void IsValidPlayerMatchP1ReturnTrue()
        {
            var p1 = Guid.NewGuid();
            var target = TestTargetInitializer.BuildGame(p1, Guid.NewGuid(), Guid.NewGuid());

            var actual = target.IsValidPlayer(p1);

            Assert.IsTrue(actual);
        }

        [TestMethod]
        public void IsValidPlayerMatchP2ReturnTrue()
        {
            var p2 = Guid.NewGuid();
            var target = TestTargetInitializer.BuildGame(Guid.NewGuid(), p2, Guid.NewGuid());

            var actual = target.IsValidPlayer(p2);

            Assert.IsTrue(actual);
        }

        [TestMethod]
        public void IsValidPlayerDoesNotMatchPlayersReturnFalse()
        {
            var target = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());

            var actual = target.IsValidPlayer(Guid.NewGuid());

            Assert.IsFalse(actual);
        }

        #endregion

        #region OnPlayerDisconnected tests

        [TestMethod]
        public void OnPlayerDisconnectedFirstPlayerDisconectedSetOpponentDisconnectedGameStatus()
        {
            var statisticRepoStub = new Mock<IStatisitcsRepository>();
            var target = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), statisticRepoStub);

            target.OnPlayerDisconnected(target.Player1Id);
            var actual = target.GameStatus;

            Assert.AreEqual(GameStatus.OpponentDisconnected, actual);
        }

        [TestMethod]
        public void OnPlayerDisconnectedSecondPlayerDisconectedSetOpponentDisconnectedGameStatus()
        {
            var statisticRepoStub = new Mock<IStatisitcsRepository>();
            var target = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), statisticRepoStub);

            target.OnPlayerDisconnected(target.Player2Id);
            var actual = target.GameStatus;

            Assert.AreEqual(GameStatus.OpponentDisconnected, actual);
        }

        [TestMethod]
        [ExpectedException(typeof(ArgumentException))]
        public void OnPlayerDisconnectedUnknownDisconectedSetOpponentDisconnectedGameStatus()
        {
            var statisticRepoStub = new Mock<IStatisitcsRepository>();
            var target = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), statisticRepoStub);

            target.OnPlayerDisconnected(Guid.NewGuid());
        }

        [TestMethod]
        public void OnPlayerDisconnectedGameInProgressStatisticsChanged()
        {
            var statisticRepoStub = new Mock<IStatisitcsRepository>();
            var target = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), statisticRepoStub);

            target.OnPlayerDisconnected(target.Player2Id);
            
            statisticRepoStub.Verify(x => x.AddLeaves(target.Player2Id, 1));
            statisticRepoStub.Verify(x => x.AddWonGames(target.Player1Id, 1));
        }

        [TestMethod]
        public void OnPlayerDisconnectedGameNotInProgressStatisticsChanged()
        {
            var statisticRepoStub = new Mock<IStatisitcsRepository>();
            var target = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid(), statisticRepoStub);
            
            target.GameStatus = GameStatus.PlayerTwoWin;
            
            target.OnPlayerDisconnected(target.Player2Id);

            statisticRepoStub.Verify(x => x.AddLeaves(target.Player2Id, 1), Times.Never);
            statisticRepoStub.Verify(x => x.AddWonGames(target.Player1Id, 1), Times.Never);
        }

        #endregion
    }
}
