﻿using System;

using TicTacToe.Model.Interfaces;

namespace TicTacToe.Model.Model
{
    public class Game
    {
        private readonly IStatisitcsRepository statisticsRepository;
        
        public Game(Guid player1, Guid player2, Guid gameId, IStatisitcsRepository statisitcsRepository)
        {
            if (player1 == Guid.Empty)
            {
                throw new ArgumentException("Player1 id can not be empty", "player1");
            }

            if (player2 == Guid.Empty)
            {
                throw new ArgumentException("Player2 id can not be empty", "player2");
            }

            if (gameId == Guid.Empty)
            {
                throw new ArgumentException("Game id can not be empty", "gameId");
            }

            this.Player1Id = player1;
            this.Player2Id = player2;
            this.GameId = gameId;
            this.NextStepPlayerId = player1;
            this.CreatedDateTime = DateTime.Now;

            this.Field = new Field();
            this.GameStatus = GameStatus.PlayerOneTurn;

            this.statisticsRepository = statisitcsRepository;
        }

        public int? NextStepOuterField { get; set; }

        public Guid Player1Id { get; set; }

        public Guid Player2Id { get; set; }

        public Guid GameId { get; set; }

        public DateTime CreatedDateTime { get; set; }

        public Guid? LastPlayerStep { get; private set; }

        public int? LastOuterField { get; private set; }

        public int? LastInnerField { get; private set; }

        public GameStatus GameStatus { get; set; }

        /// <summary>
        /// Gets or sets id of player who should make next step
        /// </summary>
        public Guid NextStepPlayerId { get; set; }

        public Field Field { get; set; }

        public GameStepStatus MakeStep(Guid playerId, int outerField, int innerField)
        {
            if (this.GameStatus == GameStatus.OpponentDisconnected)
            {
                return GameStepStatus.OpponentDisconnected;
            }

            if (playerId == Player1Id)
            {
                return MakeXStep(playerId, outerField, innerField);
            }
            
            return MakeOStep(playerId, outerField, innerField);
        }

        public bool IsAllowedStep(int outerFiled)
        {
            if (!NextStepOuterField.HasValue)
            {
                return true;
            }

            // allow player make step in to different outer field if nextStepOuterField has no free space
            if (Field[NextStepOuterField.Value].HasNoFreeSpace)
            {
                return true;
            }

            return outerFiled == NextStepOuterField;
        }

        public bool IsValidPlayer(Guid playerId)
        {
            return playerId == Player1Id || playerId == Player2Id;
        }

        public OuterField GetLastOuterField()
        {
            return LastOuterField != null ? Field[(int)LastOuterField] : null;
        }

        public void OnPlayerDisconnected(Guid disconnectedPlayerId)
        {
            if (this.IsGameInProgress())
            {
                this.statisticsRepository.AddLeaves(disconnectedPlayerId, 1);
                var winner = this.GetOpponentId(disconnectedPlayerId);
                this.statisticsRepository.AddWonGames(winner, 1);
            }

            if (disconnectedPlayerId == this.Player1Id || disconnectedPlayerId == this.Player2Id)
            {
                this.GameStatus = GameStatus.OpponentDisconnected;
            }
            else
            {
                throw new ArgumentException("Wrong player disconnected");
            }
        }

        public GameState GetGameState()
        {
            return new GameState
            {
                LastPlayerStep = this.LastPlayerStep ?? Guid.Empty,
                LastInnerField = this.LastInnerField ?? 0,
                LastOuterField = this.LastOuterField ?? 0,
                GameStatus = this.GameStatus,
                LastOuterFieldState = this.GetLastOuterField().State
            };
        }

        private bool IsGameInProgress()
        {
            return this.GameStatus == GameStatus.PlayerOneTurn || this.GameStatus == GameStatus.PlayerTwoTurn;
        }

        private GameStepStatus MakeXStep(Guid player, int outerFieldId, int innerFieldId)
        {
            return MakeStep(player, outerFieldId, innerFieldId, true);
        }

        private GameStepStatus MakeOStep(Guid player, int outerFieldId, int innerFieldId)
        {
            return MakeStep(player, outerFieldId, innerFieldId, false);
        }

        private GameStepStatus MakeStep(Guid player, int outerFieldId, int innerFieldId, bool isX)
        {
            if (!IsAllowedStep(outerFieldId))
            {
                return GameStepStatus.WrongOuterField;
            }

            if (!Field[outerFieldId].IsAllowedStep(innerFieldId))
            {
                return GameStepStatus.AlreadyFilled;
            }

            if (isX)
            {
                Field[outerFieldId, innerFieldId].MakeX();
            }
            else
            {
                Field[outerFieldId, innerFieldId].MakeO();
            }

            if (Field[outerFieldId].IsEmpty &&
                Field[outerFieldId].HasCompletedLine(innerFieldId, isX))
            {
                if (isX)
                {
                    Field[outerFieldId].MakeX();
                }
                else
                {
                    Field[outerFieldId].MakeO();
                }
            }
            else if (Field[outerFieldId].IsEmpty &&
                Field[outerFieldId].HasNoFreeSpace)
            {
                Field[outerFieldId].MakeTied();
            }

            NextStepOuterField = Field[innerFieldId].HasNoFreeSpace ? (int?)null : innerFieldId;
            SwitchNextPlayerStep();
            LastPlayerStep = player;
            LastInnerField = innerFieldId;
            LastOuterField = outerFieldId;

            if (Field.HasCompletedLine(outerFieldId, isX))
            {
                GameStatus = player == Player1Id ? GameStatus.PlayerOneWin : GameStatus.PlayerTwoWin;
                
                var opponent = this.GetOpponentId(player);
                this.statisticsRepository.AddWonGames(player, 1);
                this.statisticsRepository.AddLostGames(opponent, 1);

                return GameStepStatus.Win;
            }

            if (!Field.CanSomeoneWin())
            {
                GameStatus = GameStatus.Draw;
                this.statisticsRepository.AddDrawGames(this.Player1Id, 1);
                this.statisticsRepository.AddDrawGames(this.Player2Id, 1);

                return GameStepStatus.Draw;
            }

            GameStatus = NextStepPlayerId == Player1Id ? GameStatus.PlayerOneTurn : GameStatus.PlayerTwoTurn;
            return GameStepStatus.Ok;
        }

        private void SwitchNextPlayerStep()
        {
            NextStepPlayerId = NextStepPlayerId == Player1Id ? Player2Id : Player1Id;
        }

        private Guid GetOpponentId(Guid currentPlayerId)
        {
            return this.Player1Id == currentPlayerId ? this.Player2Id : this.Player1Id;
        }
    }
}
