﻿using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;

using TicTacToe.Common.Log;
using TicTacToe.Common.Security;
using TicTacToe.Model.Interfaces;
using TicTacToe.Model.Model;

namespace TicTacToe.Model.Repositories
{
    public class SessionRepository : ISessionRepository
    {
        private readonly ILoggerFacade logger;
        private readonly IUserRepository userRepository;
        private readonly IHashGenerator hashGenerator;

        public SessionRepository(IUserRepository userRepository = null)
        {
            this.hashGenerator = new BCryptHashGenerator();
            logger = new LoggerFacade(this.GetType());
            this.userRepository = userRepository ?? new UserRepository();
        }

        public Session Login(string login, string password)
        {
            using (var dbContext = new TicTacToeContext())
            {
                try
                {
                    var user = userRepository.GetUserByEmail(login);
                    if (!this.hashGenerator.IsValid(password, user.Password))
                    {
                        return null;
                    }

                    var session = new Session
                    {
                        Id = Guid.NewGuid(),
                        SessionToken = Guid.NewGuid().ToString() + Guid.NewGuid(), // add encryption
                        ExpirationDate = DateTime.Now.AddDays(2),
                        UserId = user.Id
                    };

                    dbContext.Sessions.Add(session);
                    dbContext.SaveChanges();

                    return session;
                }
                catch (Exception ex)
                {
                    logger.Error(ex.Message);
                    throw;
                }
            }
        }

        public Session CreateTempSession(Guid tempUserId)
        {
            using (var dbContext = new TicTacToeContext())
            {
                try
                {
                    var session = new Session
                    {
                        Id = Guid.NewGuid(),
                        SessionToken = Guid.NewGuid().ToString() + Guid.NewGuid(), // add encryption
                        ExpirationDate = DateTime.Now.AddDays(2),
                        UserId = tempUserId
                    };

                    dbContext.Sessions.Add(session);
                    dbContext.SaveChanges();

                    return session;
                }
                catch (Exception ex)
                {
                    logger.Error(ex.Message);
                    throw;
                }
            }
        }

        public IEnumerable<Session> GetAllSessions()
        {
            using (var dbContext = new TicTacToeContext())
            {
                return dbContext.Sessions.Select(x => x);
            }
        }

        public Session GetSessionById(Guid sessionId)
        {
            using (var dbContext = new TicTacToeContext())
            {
                var session = dbContext.Sessions.FirstOrDefault(item => item.Id == sessionId);
                if (session == null)
                {
                    throw new ObjectNotFoundException("Session not found. Id = " + sessionId);
                }

                return session;
            }
        }

        public Session GetSessionByUserId(Guid userId)
        {
            using (var dbContext = new TicTacToeContext())
            {
                var session = dbContext.Sessions.OrderBy(x => x.ExpirationDate).FirstOrDefault(item => item.UserId == userId);
                if (session == null)
                {
                    throw new ObjectNotFoundException("Session not found. Id = " + userId);
                }

                return session;
            }
        }

        public void Logout(Guid sessionId)
        {
            using (var dbContext = new TicTacToeContext())
            {
                var session = dbContext.Sessions.FirstOrDefault(s => s.Id == sessionId);
                if (session != null)
                {
                    dbContext.Sessions.Remove(session);
                    dbContext.SaveChanges();
                }
            }
        }

        public bool IsSessionValid(Guid sessionId)
        {
            using (var dbContext = new TicTacToeContext())
            {
                var session = dbContext.Sessions.FirstOrDefault(s => s.Id == sessionId);
                if (session != null)
                {
                    return DateTime.Now <= session.ExpirationDate;
                }

                return false;
            }
        }
    }
}
