﻿using System.Threading;

using Microsoft.VisualStudio.TestTools.UnitTesting;
using Moq;

using System;
using TicTacToe.Model.Interfaces;
using TicTacToe.Model.Model;
using TicTacToe.Service.DataContracts;
using TicTacToe.UnitTest.Helpers;

namespace TicTacToe.Service.UnitTest.TicTacToeServiceTests
{
    [TestClass]
    public class ConnectToFriendGameTest
    {
        [TestMethod, TestCategory("Service")]
        public void ConnectToFriendGameTestHasNoInitiatedGame()
        {
            var target = TestTargetInitializer.BuildTicTacToeService();

            var actual = target.ConnectToFriendGame(Guid.NewGuid());

            Assert.AreEqual(InitGameStatus.FriendGameNotFound, actual.Status);
        }

        [TestMethod, TestCategory("Service")]
        public void ConnectToFriendGameFirstPlayerDisconnected()
        {
            var target = TestTargetInitializer.BuildTicTacToeService(playerDisconnectionTimeout: 1);

            var initInfo = target.InitFriendGame();
            Thread.Sleep(2000);
            var actual = target.ConnectToFriendGame(initInfo.GameId);

            Assert.AreEqual(InitGameStatus.FriendDisconnected, actual.Status);
        }

        [TestMethod, TestCategory("Service")]
        public void ConnectToFriendGameSessionIdIsNotSpecifiedNewSessionCreated()
        {
            var repository = new Mock<ISessionRepository>();
            var target = TestTargetInitializer.BuildTicTacToeService(sessionRepo: repository);

            var initInfo = target.InitFriendGame();
            target.ConnectToFriendGame(initInfo.GameId);
            
            repository.Verify(x => x.CreateTempSession(It.IsAny<Guid>()));
        }

        [TestMethod, TestCategory("Service")]
        public void ConnectToFriendGameSessionIdIsNotSpecifiedNewSessionIdReturned()
        {
            var repository = new Mock<ISessionRepository>();
            var target = TestTargetInitializer.BuildTicTacToeService(sessionRepo: repository);

            var info = target.InitFriendGame();
            var actual = target.ConnectToFriendGame(info.GameId);

            Assert.AreNotEqual(Guid.Empty, actual.SessionId);
        }

        [TestMethod, TestCategory("Service")]
        public void ConnectToFriendGameSessionIdIsSpecifiedButNotValidSessionNotValidStatusReturned()
        {
            var repository = new Mock<ISessionRepository>();
            repository.Setup(x => x.IsSessionValid(It.IsAny<Guid>())).Returns(false);

            var target = TestTargetInitializer.BuildTicTacToeService(sessionRepo: repository);
            
            var info = target.InitFriendGame();
            var actual = target.ConnectToFriendGame(info.GameId, Guid.NewGuid());

            Assert.AreEqual(InitGameStatus.SessionNotValid, actual.Status);
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameSecondUserCallSessionIdIsSpecifiedExistingUserUsed()
        {
            var sessionId = Guid.NewGuid();
            var userId = Guid.NewGuid();
            var repository = new Mock<ISessionRepository>();
            repository.Setup(x => x.IsSessionValid(It.IsAny<Guid>())).Returns(true);
            repository.Setup(x => x.GetSessionById(It.Is<Guid>(y => y == sessionId)))
                .Returns(new Session { Id = sessionId, UserId = userId });

            var target = TestTargetInitializer.BuildTicTacToeService(sessionRepo: repository);

            var info = target.InitFriendGame();
            var actual = target.ConnectToFriendGame(info.GameId, sessionId);

            Assert.AreEqual(userId, actual.PlayerId);
            Assert.AreEqual(sessionId, actual.SessionId);
        }

        [TestMethod, TestCategory("Service")]
        public void ConnectToFriendGameSuccessfullyConnectedReturnGameInitiated()
        {
            var target = TestTargetInitializer.BuildTicTacToeService();

            var initInfo = target.InitFriendGame();
            var actual = target.ConnectToFriendGame(initInfo.GameId);

            Assert.AreEqual(InitGameStatus.GameInitiated, actual.Status);
        }

        [TestMethod, TestCategory("Service")]
        public void ConnectToFriendGameSuccessfullyConnectedGameRemovedFromQueue()
        {
            var target = TestTargetInitializer.BuildTicTacToeService();

            var initInfo = target.InitFriendGame();
            target.ConnectToFriendGame(initInfo.GameId);
            var actual = target.ConnectToFriendGame(initInfo.GameId);

            Assert.AreEqual(InitGameStatus.FriendGameNotFound, actual.Status);
        }
    }
}
