﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using Moq;

using System;
using TicTacToe.Model.Interfaces;
using TicTacToe.Model.Model;
using TicTacToe.Service.DataContracts;
using TicTacToe.UnitTest.Helpers;

namespace TicTacToe.Service.UnitTest.TicTacToeServiceTests
{
    [TestClass]
    public class InitGameTest
    {
        [TestMethod, TestCategory("Service")]
        public void InitGameFirstPlayerConnectedReturnWaitForOpponentStatus()
        {
            var target = TestTargetInitializer.BuildTicTacToeService();

            var actual = target.InitGame();

            Assert.AreEqual(InitGameStatus.WaitForOpponent, actual.Status);
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameFirstPlayerConnectedGameIdSet()
        {
            var target = TestTargetInitializer.BuildTicTacToeService();

            var actual = target.InitGame();

            Assert.AreNotEqual(Guid.Empty, actual.GameId);
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameFirstPlayerConnectedPlayerIdSet()
        {
            var target = TestTargetInitializer.BuildTicTacToeService();

            var actual = target.InitGame();

            Assert.AreNotEqual(Guid.Empty, actual.PlayerId);
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameFirstPlayerConnectedIsYourStepFirstTrue()
        {
            var target = TestTargetInitializer.BuildTicTacToeService();

            var actual = target.InitGame();

            Assert.IsTrue(actual.IsYourStepFirst);
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameSecondPlayerConnectedReturnGameInitiatedStatus()
        {
            var target = TestTargetInitializer.BuildTicTacToeService();

            target.InitGame();
            var actualSecondPlayer = target.InitGame();

            Assert.AreEqual(InitGameStatus.GameInitiated, actualSecondPlayer.Status);
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameSecondPlayerConnectedBothPlayerHasSameGameId()
        {
            var target = TestTargetInitializer.BuildTicTacToeService();

            var actualFirstPlayer = target.InitGame();
            var actualSecondPlayer = target.InitGame();

            Assert.AreEqual(actualFirstPlayer.GameId, actualSecondPlayer.GameId);
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameThirdPlayerConnectedReturnWaitForOpponentStatus()
        {
            var target = TestTargetInitializer.BuildTicTacToeService();

            target.InitGame();
            target.InitGame();
            var actualThirdPlayer = target.InitGame();

            Assert.AreEqual(InitGameStatus.WaitForOpponent, actualThirdPlayer.Status);
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameGameInitiatedGameIdSet()
        {
            var target = TestTargetInitializer.BuildTicTacToeService();

            target.InitGame();
            var actual = target.InitGame();
            
            Assert.AreNotEqual(Guid.Empty, actual.GameId);
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameGameInitiatedPlayerIdSet()
        {
            var target = TestTargetInitializer.BuildTicTacToeService();

            target.InitGame();
            var actual = target.InitGame();

            Assert.AreNotEqual(Guid.Empty, actual.PlayerId);
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameGameInitiatedIsYourStepFirstFalse()
        {
            var target = TestTargetInitializer.BuildTicTacToeService();

            target.InitGame();
            var actual = target.InitGame();

            Assert.IsFalse(actual.IsYourStepFirst);
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameTwoPlayersInitGameAddedToRepository()
        {
            var repository = new Mock<IGamesRepository>();
            var target = TestTargetInitializer.BuildTicTacToeService(repository.Object);

            target.InitGame();
            var actual = target.InitGame();

            repository.Verify(item => item.AddGame(It.Is<Game>(x => x.GameId == actual.GameId)));
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameSessionIdIsNotSpecifiedNewSessionCreated()
        {
            var repository = new Mock<ISessionRepository>();
            var target = TestTargetInitializer.BuildTicTacToeService(sessionRepo: repository);

            target.InitGame();
            
            repository.Verify(x => x.CreateTempSession(It.IsAny<Guid>()));
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameSessionIdIsNotSpecifiedNewSessionIdReturned()
        {
            var repository = new Mock<ISessionRepository>();
            var target = TestTargetInitializer.BuildTicTacToeService(sessionRepo: repository);

            var actual = target.InitGame();

            Assert.AreNotEqual(Guid.Empty, actual.SessionId);
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameSessionIdIsSpecifiedButNotValidSessionNotValidStatusReturned()
        {
            var repository = new Mock<ISessionRepository>();
            repository.Setup(x => x.IsSessionValid(It.IsAny<Guid>())).Returns(false);

            var target = TestTargetInitializer.BuildTicTacToeService(sessionRepo: repository);

            var actual = target.InitGame(Guid.NewGuid());

            Assert.AreEqual(InitGameStatus.SessionNotValid, actual.Status);
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameSessionIdIsSpecifiedExistingUserUsed()
        {
            var sessionId = Guid.NewGuid();
            var userId = Guid.NewGuid();
            var repository = new Mock<ISessionRepository>();
            repository.Setup(x => x.IsSessionValid(It.IsAny<Guid>())).Returns(true);
            repository.Setup(x => x.GetSessionById(It.Is<Guid>(y => y == sessionId)))
                .Returns(new Session { Id = sessionId, UserId = userId });

            var target = TestTargetInitializer.BuildTicTacToeService(sessionRepo: repository);

            var actual = target.InitGame(sessionId);

            Assert.AreEqual(userId, actual.PlayerId);
            Assert.AreEqual(sessionId, actual.SessionId);
            repository.Verify(x => x.CreateTempSession(It.IsAny<Guid>()), Times.Never());
        }

        [TestMethod, TestCategory("Service")]
        public void InitGameSecondUserCallSessionIdIsSpecifiedExistingUserUsed()
        {
            var sessionId = Guid.NewGuid();
            var userId = Guid.NewGuid();
            var repository = new Mock<ISessionRepository>();
            repository.Setup(x => x.IsSessionValid(It.IsAny<Guid>())).Returns(true);
            repository.Setup(x => x.GetSessionById(It.Is<Guid>(y => y == sessionId)))
                .Returns(new Session { Id = sessionId, UserId = userId });

            var target = TestTargetInitializer.BuildTicTacToeService(sessionRepo: repository);

            target.InitGame();
            var actual = target.InitGame(sessionId);

            Assert.AreEqual(userId, actual.PlayerId);
            Assert.AreEqual(sessionId, actual.SessionId);
        }
    }
}
