﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using System;

using Moq;

using TicTacToe.Common.Log;
using TicTacToe.Model.Interfaces;
using TicTacToe.Model.Model;
using TicTacToe.Model.Repositories;
using TicTacToe.Service.DataContracts;
using TicTacToe.Service.Interfaces;
using TicTacToe.UnitTest.Helpers;

namespace TicTacToe.Service.UnitTest.TicTacToeServiceTests
{
    [TestClass]
    public class MakeStepTest
    {
        [TestMethod]
        public void MakeStepNoGamesTest()
        {
            var games = new InMemoryGamesRepository();

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var game = new GameAndPlayerInfo() { PlayerId = Guid.NewGuid(), GameId = Guid.NewGuid() };
            
            var actual = service.MakeStep(game, 0, 0);

            Assert.AreEqual(StepStatus.WrongGameId, actual.StepStatus);
        }

        [TestMethod]
        public void MakeStepNoPlayerTest()
        {
            var gameId = Guid.NewGuid();

            var games = new InMemoryGamesRepository();
            games.AddGame(TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), gameId));

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var game = new GameAndPlayerInfo() { PlayerId = Guid.NewGuid(), GameId = gameId };

            var actual = service.MakeStep(game, 0, 0);

            Assert.AreEqual(StepStatus.WrongPlayerId, actual.StepStatus);
        }

        [TestMethod]
        public void MakeStepWrongPlayerTest()
        {
            var gameId = Guid.NewGuid();
            var player1 = Guid.NewGuid();
            var player2 = Guid.NewGuid();

            var games = new InMemoryGamesRepository();
            games.AddGame(TestTargetInitializer.BuildGame(player1, player2, gameId));

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var game = new GameAndPlayerInfo() { PlayerId = player2, GameId = gameId };

            var actual = service.MakeStep(game, 0, 0);

            Assert.AreEqual(StepStatus.WrongPlayerMakeStep, actual.StepStatus);
        }

        [TestMethod]
        public void MakeStepWrongStepInNonEmptyFieldTest()
        {
            var games = new InMemoryGamesRepository();
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            game.Field[0][0].MakeO();
            games.AddGame(game);

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var firstStep = new GameAndPlayerInfo() { PlayerId = game.Player1Id, GameId = game.GameId };
            
            var actual = service.MakeStep(firstStep, 0, 0);

            Assert.AreEqual(StepStatus.WrongInnerFieldStep, actual.StepStatus);
        }

        [TestMethod]
        public void MakeStepFirstStepSuccessTest()
        {
            var games = new InMemoryGamesRepository();
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            games.AddGame(game);

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var step = new GameAndPlayerInfo() { PlayerId = game.Player1Id, GameId = game.GameId };

            var actual = service.MakeStep(step, 0, 0);

            Assert.AreEqual(StepStatus.WaitForNextPlayerStep, actual.StepStatus);
        }

        [TestMethod]
        public void MakeStepFirstStepSuccessFieldStateTest()
        {
            var games = new InMemoryGamesRepository();
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            games.AddGame(game);

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var step = new GameAndPlayerInfo() { PlayerId = game.Player1Id, GameId = game.GameId };

            var actual = service.MakeStep(step, 0, 0);

            Assert.AreEqual(StepStatus.WaitForNextPlayerStep, actual.StepStatus);

            for (int i = 0; i < 9; i++)
            {
                for (int j = 0; j < 9; j++)
                {
                    if (i == 0 && j == 0)
                    {
                        Assert.IsTrue(game.Field[i, j].IsX);
                        continue;
                    }

                    Assert.IsTrue(game.Field[i, j].IsEmpty);
                }
            }
        }

        [TestMethod]
        public void MakeStepFirstStepSuccessNextPlayerIdTest()
        {
            var games = new InMemoryGamesRepository();
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            games.AddGame(game);

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var step = new GameAndPlayerInfo() { PlayerId = game.Player1Id, GameId = game.GameId };

            var actual = service.MakeStep(step, 0, 0);

            Assert.AreEqual(game.Player2Id, game.NextStepPlayerId);
        }

        [TestMethod]
        public void MakeStepSecondStepSuccessNextPlayerIdTest()
        {
            var games = new InMemoryGamesRepository();
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            games.AddGame(game);

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var firstStep = new GameAndPlayerInfo() { PlayerId = game.Player1Id, GameId = game.GameId };
            var secondStep = new GameAndPlayerInfo() { PlayerId = game.Player2Id, GameId = game.GameId };

            var actual = service.MakeStep(firstStep, 0, 0);

            Assert.AreEqual(StepStatus.WaitForNextPlayerStep, actual.StepStatus);

            actual = service.MakeStep(secondStep, 0, 1);

            Assert.AreEqual(game.Player1Id, game.NextStepPlayerId);
        }
        
        [TestMethod]
        public void MakeStepSecondStepSuccessFieldStateTest()
        {
            var games = new InMemoryGamesRepository();
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            games.AddGame(game);

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var firstStep = new GameAndPlayerInfo() { PlayerId = game.Player1Id, GameId = game.GameId };
            var secondStep = new GameAndPlayerInfo() { PlayerId = game.Player2Id, GameId = game.GameId };

            var actual = service.MakeStep(firstStep, 0, 0);

            Assert.AreEqual(StepStatus.WaitForNextPlayerStep, actual.StepStatus);

            actual = service.MakeStep(secondStep, 0, 1);

            for (int i = 0; i < 9; i++)
            {
                for (int j = 0; j < 9; j++)
                {
                    if (i == 0 && j == 0)
                    {
                        Assert.IsTrue(game.Field[i, j].IsX);
                        continue;
                    }

                    if (i == 0 && j == 1)
                    {
                        Assert.IsTrue(game.Field[i, j].IsO);
                        continue;
                    }

                    Assert.IsTrue(game.Field[i, j].IsEmpty);
                }
            }
        }

        [TestMethod]
        public void MakeStepSecondStepFailWrongInnerIdTest()
        {
            var games = new InMemoryGamesRepository();
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            games.AddGame(game);

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var firstStep = new GameAndPlayerInfo() { PlayerId = game.Player1Id, GameId = game.GameId };
            var secondStep = new GameAndPlayerInfo() { PlayerId = game.Player2Id, GameId = game.GameId };

            var actual = service.MakeStep(firstStep, 0, 0);

            Assert.AreEqual(StepStatus.WaitForNextPlayerStep, actual.StepStatus);

            actual = service.MakeStep(secondStep, 0, 0);

            Assert.AreEqual(StepStatus.WrongInnerFieldStep, actual.StepStatus);
        }

        [TestMethod]
        public void MakeStepSecondStepFailWrongOuterIdTest()
        {
            var games = new InMemoryGamesRepository();
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            games.AddGame(game);

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var firstStep = new GameAndPlayerInfo() { PlayerId = game.Player1Id, GameId = game.GameId };
            var secondStep = new GameAndPlayerInfo() { PlayerId = game.Player2Id, GameId = game.GameId };

            var actual = service.MakeStep(firstStep, 0, 0);

            Assert.AreEqual(StepStatus.WaitForNextPlayerStep, actual.StepStatus);

            actual = service.MakeStep(secondStep, 1, 0);

            Assert.AreEqual(StepStatus.WrongOuterFieldStep, actual.StepStatus);
        }

        [TestMethod]
        public void MakeStepInnerFieldHasNoFreeSpaceSuccess()
        {
            var games = new InMemoryGamesRepository();
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            for (int i = 0; i < 8; i++)
            {
                game.Field[0][i].MakeX();
            }

            games.AddGame(game);

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var firstStep = new GameAndPlayerInfo() { PlayerId = game.Player1Id, GameId = game.GameId };
            var secondStep = new GameAndPlayerInfo() { PlayerId = game.Player2Id, GameId = game.GameId };
            var thirdStep = new GameAndPlayerInfo() { PlayerId = game.Player1Id, GameId = game.GameId };

            // send second player to the right bottom corner. after this step left upper field completely filled.
            var actual = service.MakeStep(firstStep, 0, 8);

            Assert.AreEqual(StepStatus.WaitForNextPlayerStep, actual.StepStatus);

            // send first palyer to the left upper field. as this field has no free space, player 1 can make step in any free field.
            actual = service.MakeStep(secondStep, 8, 0);

            Assert.AreEqual(StepStatus.WaitForNextPlayerStep, actual.StepStatus);

            // first player can make step in any free field.
            actual = service.MakeStep(thirdStep, 1, 1);

            Assert.AreEqual(StepStatus.WaitForNextPlayerStep, actual.StepStatus);
        }

        [TestMethod]
        public void MakeStepOuterFieldSetToXAndCompletelyFilledRemainInXState()
        {
            var games = new InMemoryGamesRepository();
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            game.Field[0].Fill(@"xoo
                                 oxo
                                 x--");
            games.AddGame(game);

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var firstStep = new GameAndPlayerInfo() { PlayerId = game.Player1Id, GameId = game.GameId };
            var secondStep = new GameAndPlayerInfo() { PlayerId = game.Player2Id, GameId = game.GameId };
            
            var actual = service.MakeStep(firstStep, 0, 8);

            Assert.AreEqual(StepStatus.WaitForNextPlayerStep, actual.StepStatus);
            Assert.AreEqual(OuterFieldState.X, game.Field[0].State);

            actual = service.MakeStep(secondStep, 8, 0);
            actual = service.MakeStep(firstStep, 0, 7);

            Assert.AreEqual(StepStatus.WaitForNextPlayerStep, actual.StepStatus);
            Assert.AreEqual(OuterFieldState.X, game.Field[0].State);
        }

        [TestMethod]
        public void MakeStepOuterFieldSetToXAndOHasComletedLineFieldRemainInXState()
        {
            var games = new InMemoryGamesRepository();
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            game.Field[0].Fill(@"xx-
                                 o-o
                                 ---");
            games.AddGame(game);

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var firstStep = new GameAndPlayerInfo() { PlayerId = game.Player1Id, GameId = game.GameId };
            var secondStep = new GameAndPlayerInfo() { PlayerId = game.Player2Id, GameId = game.GameId };

            var actual = service.MakeStep(firstStep, 0, 2);

            Assert.AreEqual(StepStatus.WaitForNextPlayerStep, actual.StepStatus);
            Assert.AreEqual(OuterFieldState.X, game.Field[0].State);

            actual = service.MakeStep(secondStep, 2, 1);
            actual = service.MakeStep(firstStep, 1, 0);
            actual = service.MakeStep(secondStep, 0, 4);

            Assert.AreEqual(StepStatus.WaitForNextPlayerStep, actual.StepStatus);
            Assert.AreEqual(OuterFieldState.X, game.Field[0].State);
        }

        [TestMethod]
        public void MakeStepXMakeStepWinTheGame()
        {
            var games = new InMemoryGamesRepository();
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            game.Field.Fill(@"xx-
                              o-o
                              ---");

            game.Field[2].Fill(@"x-x
                                 -o-
                                 ---");
            games.AddGame(game);

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var firstStep = new GameAndPlayerInfo() { PlayerId = game.Player1Id, GameId = game.GameId };
            
            var actual = service.MakeStep(firstStep, 2, 1);

            Assert.AreEqual(StepStatus.YouWin, actual.StepStatus);
        }

        [TestMethod]
        public void MakeStepOMakeStepWinTheGame()
        {
            var games = new InMemoryGamesRepository();
            var game = TestTargetInitializer.BuildGame(Guid.NewGuid(), Guid.NewGuid(), Guid.NewGuid());
            game.Field.Fill(@"xx-
                              o-o
                              ---");

            game.Field[4].Fill(@"o-o
                                 -x-
                                 ---");
            games.AddGame(game);

            var service = TestTargetInitializer.BuildTicTacToeService(games);

            var firstStep = new GameAndPlayerInfo() { PlayerId = game.Player1Id, GameId = game.GameId };
            var secondStep = new GameAndPlayerInfo() { PlayerId = game.Player2Id, GameId = game.GameId };

            var actual = service.MakeStep(firstStep, 0, 4);
            Assert.AreEqual(StepStatus.WaitForNextPlayerStep, actual.StepStatus);

            actual = service.MakeStep(secondStep, 4, 1);
            Assert.AreEqual(StepStatus.YouWin, actual.StepStatus);
        }
    }
}

