﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using TicTacToe.Common.Log;
using TicTacToe.Service.Interfaces;

namespace TicTacToe.Service.Communication
{
    public sealed class PlayerConnectionPool : IConnectionPool, IDisposable
    {
        private bool isStarted;
        CancellationTokenSource cts = new CancellationTokenSource();
        CancellationToken token;
        ILoggerFacade logger;
        TimeSpan disconnectionTimeout = TimeSpan.FromSeconds(10);

        Dictionary<Guid, IConnection> connections = new Dictionary<Guid, IConnection>();

        public PlayerConnectionPool(ILoggerFacade logger = null)
        {
            this.logger = logger ?? new LoggerFacade(this.GetType());
        }
        
        public void Start()
        {
            if (this.isStarted)
            {
                return;
            }

            Task.Factory.StartNew(Scheduler, token);
            this.isStarted = true;
            this.logger.Info("Connection pool started");
        }

        public void Stop()
        {
            if (!this.isStarted)
            {
                return;
            }

            cts.Cancel();
            this.isStarted = false;
            this.logger.Info("Connection pool stoped");
        }

        public void AddConnection(IConnection connection)
        {
            if (this.connections.ContainsKey(connection.Id))
            {
                this.logger.Warn("Connection pool already contains connection with id " + connection.Id);
                return;
            }

            this.connections.Add(connection.Id, connection);
            this.logger.Debug(string.Format("Connection {0} was added into connection pool", connection.Id));
        }

        public void ConnectionStillActive(Guid id)
        {
            if (!this.connections.ContainsKey(id))
            {
                this.logger.Warn("Connection pool doesn't contains connection with id " + id);
                return;
            }

            this.connections[id].MessageArrived();
        }

        public void Dispose()
        {
            cts.Dispose();
        }

        private void Scheduler()
        {
            while (true)
            {
                this.token.ThrowIfCancellationRequested();
                
                this.CheckConnections();

                Thread.Sleep(1000);
            }
        }

        private void CheckConnections()
        {
            this.connections.Values
                .Where(item => DateTime.Now - item.LastMessageArrived >= disconnectionTimeout)
                .ToList()
                .ForEach(this.Disconnect);
        }

        private void Disconnect(IConnection connection)
        {
            if (!this.connections.ContainsKey(connection.Id))
            {
                this.logger.Warn("Cannot disconnect. Connection pool doesn't contains connection with id " + connection.Id);
                return;
            }

            connection.Disconnect();
            this.connections.Remove(connection.Id);

            this.logger.Debug(string.Format("Connection {0} was removed from connection pool", connection.Id));
        }
    }
}
