﻿using System;
using System.Collections.Generic;
using System.ServiceModel;
using System.Threading;
using System.Threading.Tasks;
using TicTacToe.Common.Log;
using TicTacToe.Common.Security;
using TicTacToe.Model.Interfaces;
using TicTacToe.Model.Model;
using TicTacToe.Model.Repositories;
using TicTacToe.Service.Communication;
using TicTacToe.Service.DataContracts;
using TicTacToe.Service.Interfaces;

namespace TicTacToe.Service
{
    [ServiceBehavior(InstanceContextMode = InstanceContextMode.Single)]
    public sealed class TicTacToeService : ITicTacToeService, IDisposable
    {
        private const int defaultPayerDisconnectionTimeoutSeconds = 5;

        private readonly TimeSpan playerDisconnectionTimeout;
        private readonly IGamesRepository games;
        private readonly IUserRepository userRepository;
        private readonly ISessionRepository sessionRepository;
        private readonly IStatisitcsRepository statisitcsRepository;

        private readonly ILoggerFacade logger;
        private readonly IConnectionPool connectionPool;

        private readonly object syncContext1 = new object();
        private readonly Dictionary<Guid, FriendGameInitInfo> waitingForOpponentFriendGames = new Dictionary<Guid, FriendGameInitInfo>();

        private readonly IHashGenerator md5HashGenerator = new MD5HashGenerator();
        private IHashGenerator hashGenerator;

        private CancellationTokenSource cts = new CancellationTokenSource();

        private Guid p1;
        private Guid waitingForOpponentGameId;
        private DateTime lastFirstPlayerPingDate;

        public TicTacToeService()
        {
            games = new InMemoryGamesRepository();
            userRepository = new UserRepository();
            sessionRepository = new SessionRepository();
            statisitcsRepository = new StatisticsRepository();
            
            logger = new LoggerFacade(GetType());
            playerDisconnectionTimeout = TimeSpan.FromSeconds(defaultPayerDisconnectionTimeoutSeconds);
            connectionPool = new PlayerConnectionPool();
            InitService();
        }

        public TicTacToeService(
            IGamesRepository gamesRepo, 
            ILoggerFacade loggerStub, 
            IConnectionPool stubConnectionPool,
            IUserRepository stubUserRepository,
            IStatisitcsRepository stubStatisitcsRepository,
            ISessionRepository stubSessionRepository,
            int? disconectionTimeout = null)
        {
            games = gamesRepo;
            logger = loggerStub;
            playerDisconnectionTimeout = TimeSpan.FromSeconds(disconectionTimeout ?? defaultPayerDisconnectionTimeoutSeconds);
            connectionPool = stubConnectionPool;
            userRepository = stubUserRepository;
            statisitcsRepository = stubStatisitcsRepository;
            sessionRepository = stubSessionRepository;
            InitService();
        }
        
        public GameInitInfo InitGame(Guid? sessionId = null)
        {
            lock (syncContext1)
            {
                Guid userId;
                Guid returnedSessionId;
                if (!CheckSession(sessionId, out userId, out returnedSessionId))
                {
                    return new GameInitInfo { Status = InitGameStatus.SessionNotValid };
                }

                if (p1 == Guid.Empty)
                {
                    p1 = userId;
                    waitingForOpponentGameId = Guid.NewGuid();

                    cts.Dispose();
                    cts = new CancellationTokenSource();
                    lastFirstPlayerPingDate = DateTime.Now;
                    Task.Factory.StartNew(FirstPlayerAvailabilityChecker, cts.Token);
                    logger.Info("First user connected. Waiting for opponent.");
                    return new GameInitInfo
                        { 
                            Status = InitGameStatus.WaitForOpponent, 
                            GameId = waitingForOpponentGameId, 
                            PlayerId = p1,
                            IsYourStepFirst = true,
                            SessionId = returnedSessionId
                        };
                }

                logger.Info("Second user connected. Game created.");
                var firstPlayer = this.p1;
                var gameId = this.waitingForOpponentGameId;
                
                this.cts.Cancel();
                this.ResetQueue();
                
                return this.CreateGame(firstPlayer, userId, gameId, returnedSessionId);
            }
        }

        public GameInitInfo InitFriendGame(Guid? sessionId = null)
        {
            lock (syncContext1)
            {
                Guid userId;
                Guid returnedSessionId;
                if (!CheckSession(sessionId, out userId, out returnedSessionId))
                {
                    return new GameInitInfo { Status = InitGameStatus.SessionNotValid };
                }

                var gameId = Guid.NewGuid();
                waitingForOpponentFriendGames.Add(gameId, new FriendGameInitInfo(gameId, userId));

                logger.Info("User initiate friend game. Waiting for opponent.");
                return new GameInitInfo
                {
                    Status = InitGameStatus.WaitForOpponent,
                    GameId = gameId,
                    PlayerId = userId,
                    IsYourStepFirst = true,
                    SessionId = returnedSessionId
                };
            }
        }

        public GameInitInfo ConnectToFriendGame(Guid gameId, Guid? sessionId = null)
        {
            lock (syncContext1)
            {
                if (!this.waitingForOpponentFriendGames.ContainsKey(gameId))
                {
                    return new GameInitInfo { Status = InitGameStatus.FriendGameNotFound };
                }

                var gameInfo = this.waitingForOpponentFriendGames[gameId];

                if (gameInfo.IsOpponentConnected)
                {
                    return new GameInitInfo { Status = InitGameStatus.FriendGameNotFound };
                }

                if (DateTime.Now - gameInfo.LastFirstPlayerUpdateDate > playerDisconnectionTimeout)
                {
                    this.waitingForOpponentFriendGames.Remove(gameId);
                    return new GameInitInfo { Status = InitGameStatus.FriendDisconnected };
                }

                Guid userId;
                Guid returnedSessionId;
                if (!CheckSession(sessionId, out userId, out returnedSessionId))
                {
                    return new GameInitInfo { Status = InitGameStatus.SessionNotValid };
                }
                
                logger.Info("Second user connected to friend game. Game created.");
                gameInfo.OpponentConnected();
                return this.CreateGame(gameInfo.FirstPlayerId, userId, gameId, returnedSessionId);
            }
        }

        public CheckGameInitStatusResult CheckInitGameStatus(GameAndPlayerInfo gameAndPlayerInfo)
        {
            lock (syncContext1)
            {
                if (this.waitingForOpponentFriendGames.ContainsKey(gameAndPlayerInfo.GameId))
                {
                    var gameInfo = this.waitingForOpponentFriendGames[gameAndPlayerInfo.GameId];
                    if (gameInfo.FirstPlayerId != gameAndPlayerInfo.PlayerId)
                    {
                        return new CheckGameInitStatusResult { Status = CheckGameInitStatus.WrongPlayerId };
                    }

                    if (!gameInfo.IsOpponentConnected)
                    {
                        this.waitingForOpponentFriendGames[gameAndPlayerInfo.GameId].StillWaitingForOpponent();
                        return new CheckGameInitStatusResult { Status = CheckGameInitStatus.WaitForOpponent };
                    }

                    var opponentId = this.games.GetGame(gameInfo.GameId).Player2Id;
                    this.waitingForOpponentFriendGames.Remove(gameAndPlayerInfo.GameId);
                    return new CheckGameInitStatusResult { Status = CheckGameInitStatus.GameInitiated, OpponentId = opponentId };
                }

                logger.Debug("First user check game init status.");
                if (waitingForOpponentGameId == gameAndPlayerInfo.GameId && p1 == gameAndPlayerInfo.PlayerId)
                {
                    lastFirstPlayerPingDate = DateTime.Now;
                    return new CheckGameInitStatusResult { Status = CheckGameInitStatus.WaitForOpponent };
                }

                if (waitingForOpponentGameId == gameAndPlayerInfo.GameId && p1 != gameAndPlayerInfo.PlayerId)
                {
                    return new CheckGameInitStatusResult { Status = CheckGameInitStatus.WrongPlayerId };
                }

                if (waitingForOpponentGameId != gameAndPlayerInfo.GameId)
                {
                    if (!games.ContainsGame(gameAndPlayerInfo.GameId))
                    {
                        return new CheckGameInitStatusResult { Status = CheckGameInitStatus.WrongGameId };
                    }

                    var opponentId = this.games.GetGame(gameAndPlayerInfo.GameId).Player2Id;
                    return new CheckGameInitStatusResult { Status = CheckGameInitStatus.GameInitiated, OpponentId = opponentId };
                }

                return new CheckGameInitStatusResult { Status = (CheckGameInitStatus)5 };
            }
        }

        public StepInfo MakeStep(GameAndPlayerInfo gameAndPlayerInfo, int outerField, int innerField)
        {
            var game = games.GetGame(gameAndPlayerInfo.GameId);
            var playerId = gameAndPlayerInfo.PlayerId;
            var stepInfo = new StepInfo { PlayerId = playerId, InnerField = innerField, OuterField = outerField };
            if (game == null)
            {
                stepInfo.StepStatus = StepStatus.WrongGameId;
                return stepInfo;
            }

            if (!game.IsValidPlayer(playerId))
            {
                stepInfo.StepStatus = StepStatus.WrongPlayerId;
                return stepInfo;
            }

            if (game.NextStepPlayerId != gameAndPlayerInfo.PlayerId)
            {
                stepInfo.StepStatus = StepStatus.WrongPlayerMakeStep;
                return stepInfo;
            }

            GameStepStatus stepStatus = game.MakeStep(playerId, outerField, innerField);
            stepInfo.OuterFieldStatus = game.Field[outerField].State;
            stepInfo.NextStepAllowedOuterField = game.NextStepOuterField;

            if (stepStatus == GameStepStatus.AlreadyFilled)
            {
                stepInfo.StepStatus = StepStatus.WrongInnerFieldStep;
                return stepInfo;
            }

            if (stepStatus == GameStepStatus.WrongOuterField)
            {
                stepInfo.StepStatus = StepStatus.WrongOuterFieldStep;
                return stepInfo;
            }

            if (stepStatus == GameStepStatus.Win)
            {
                stepInfo.StepStatus = StepStatus.YouWin;
                return stepInfo;
            }

            if (stepStatus == GameStepStatus.Draw)
            {
                stepInfo.StepStatus = StepStatus.Draw;
                return stepInfo;
            }

            if (stepStatus == GameStepStatus.OpponentDisconnected)
            {
                RemoveGame(game.GameId);
                stepInfo.StepStatus = StepStatus.OpponentDisconnected;
                return stepInfo;
            }

            stepInfo.StepStatus = StepStatus.WaitForNextPlayerStep;
            return stepInfo;           
        }

        public StepStatus IsMyStep(GameAndPlayerInfo gameAndPlayerInfo)
        {
            var game = games.GetGame(gameAndPlayerInfo.GameId);
            var playerId = gameAndPlayerInfo.PlayerId;
            if (game == null)
            {
                return StepStatus.WaitForNextPlayerStep;
            }

            if (!game.IsValidPlayer(playerId))
            {
                return StepStatus.WaitForNextPlayerStep;
            }

            if (game.GameStatus == GameStatus.OpponentDisconnected)
            {
                RemoveGame(game.GameId);
                return StepStatus.OpponentDisconnected;
            }

            return game.NextStepPlayerId == playerId ? StepStatus.YourTurn : StepStatus.WaitForNextPlayerStep;
        }
        
        public StepInfo GetLastStep(GameAndPlayerInfo gameAndPlayerInfo)
        {
            var game = games.GetGame(gameAndPlayerInfo.GameId);
            var playerId = gameAndPlayerInfo.PlayerId;
            if (game == null)
            {
                return null;
            }

            if (!game.IsValidPlayer(playerId))
            {
                return null;
            }

            if (game.LastPlayerStep == null)
            {
                return null;
            }

            var gameState = game.GetGameState();

            if (gameState.GameStatus == GameStatus.OpponentDisconnected)
            {
                RemoveGame(game.GameId); 
            }

            return new StepInfo
            {
                PlayerId = gameState.LastPlayerStep,
                InnerField = gameState.LastInnerField,
                OuterField = gameState.LastOuterField,
                GameStatus = gameState.GameStatus,
                OuterFieldStatus = gameState.LastOuterFieldState
            };
        }

        public IEnumerable<Game> GetAllGames()
        {
            return games.GetAllGames();
        }

        public void IAmLive(GameAndPlayerInfo gameAndPlayerInfo)
        {
            var game = games.GetGame(gameAndPlayerInfo.GameId);
            var playerId = gameAndPlayerInfo.PlayerId;
            if (game == null)
            {
                logger.Warn("Player says that his is live, but not game found. " + gameAndPlayerInfo);
                return;
            }

            if (!game.IsValidPlayer(playerId))
            {
                logger.Warn("Not valid player try to say thay it is live. " + gameAndPlayerInfo);
                return;
            }

            connectionPool.ConnectionStillActive(gameAndPlayerInfo.PlayerId);
        }

        public RegistrationResult RegisterUser(UserRegistrationInfo registrationInfo)
        {
            try
            {
                var user = new User
                {
                    Id = Guid.NewGuid(),
                    Email = registrationInfo.Email,
                    Login = registrationInfo.Email,
                    Password = hashGenerator.GenerateHash(registrationInfo.Password),
                    RegisteredDate = DateTime.Now
                };

                if (userRepository.IsUserExists(registrationInfo.Email))
                {
                    return new RegistrationResult(false, "User already exists");
                }

                userRepository.CreateUser(user);

                if (registrationInfo.SessionId.HasValue
                    && this.sessionRepository.IsSessionValid(registrationInfo.SessionId.Value))
                {
                    CopyUserStatsFromTempSession(registrationInfo.SessionId.Value, user.Id);
                }

                logger.Info(string.Format("User {0} successfully registered", user.Login));
                return new RegistrationResult(true);
            }
            catch (Exception ex)
            {
                logger.Warn(ex.Message);
                //// TODO: Add user friendly error messages.
                return new RegistrationResult(false, ex.Message);
            }
        }

        public LoginResult Login(UserLoginInfo loginInfo)
        {
            try
            {
                var session = sessionRepository.Login(loginInfo.Login, loginInfo.Password);
                if (session != null)
                {
                    return new LoginResult
                    {
                        SessionId = session.Id,
                        UserId = session.UserId,
                        IsSuccess = true
                    };
                }
            }
            catch (Exception)
            {
                return new LoginResult
                {
                    IsSuccess = false
                };
            }

            return new LoginResult
            {
                IsSuccess = false
            };
        }

        public void Logout(Guid sessionId)
        {
            this.sessionRepository.Logout(sessionId);
        }

        public UserInfoResult GetUserInfo(Guid? userId, Guid sessionId)
        {
            var result = new UserInfoResult();
            
            var session = this.sessionRepository.GetSessionById(sessionId);

            try
            {
                if (userId.HasValue)
                {
                    if (this.userRepository.IsUserExists(userId.Value))
                    {
                        var user = this.userRepository.GetUserById(userId.Value);
                        result.Email = user.Email;
                    }
                    else
                    {
                        session = this.sessionRepository.GetSessionByUserId(userId.Value);
                    }
                }
                else
                {
                    result.Email = "Anonymous";
                }

                if (result.Email == null)
                {
                    result.Email = "Anonymous";
                }

                result.AvatarHash = md5HashGenerator.GenerateHash(result.Email.Trim().ToLower());

                var stats = this.statisitcsRepository.GetUserStatistics(userId ?? session.UserId);
                if (stats != null)
                {
                    result.WonGameCount = stats.WonGames;
                    result.LostGameCount = stats.LostGames;
                    result.DrawGameCount = stats.DrawGames;
                    result.LeaveGameCount = stats.Leaves;
                    result.PlayedGameCount = stats.GamesPlayed;
                }

                result.IsSuccess = true;
            }
            catch (Exception)
            {
                return result;
            }

            return result;
        }

        public void Dispose()
        {
            cts.Cancel();
            cts.Dispose();
            connectionPool.Stop();
        }

        private GameInitInfo CreateGame(Guid firstPlayerId, Guid secondPlayerId, Guid gameId, Guid sesionId)
        {
            var game = new Game(firstPlayerId, secondPlayerId, gameId, new StatisticsRepository());
            this.games.AddGame(game);

            var p1Connection = new PlayerConnection(game.Player1Id);
            var p2Connection = new PlayerConnection(game.Player2Id);
            this.connectionPool.AddConnection(p1Connection);
            this.connectionPool.AddConnection(p2Connection);
            p1Connection.PlayerDisconnected += (sender, args) => game.OnPlayerDisconnected(args.Id);
            p2Connection.PlayerDisconnected += (sender, args) => game.OnPlayerDisconnected(args.Id);
            
            this.statisitcsRepository.AddGamesPlayed(game.Player1Id, 1);
            this.statisitcsRepository.AddGamesPlayed(game.Player2Id, 1);

            return new GameInitInfo
            {
                Status = InitGameStatus.GameInitiated,
                GameId = game.GameId,
                PlayerId = game.Player2Id,
                IsYourStepFirst = false,
                SessionId = sesionId,
                OpponentId = game.Player1Id
            };
        }

        private bool CheckSession(Guid? sessionId, out Guid userId, out Guid returnedSessionId)
        {
            userId = Guid.NewGuid();
            returnedSessionId = sessionId ?? Guid.NewGuid();
            if (sessionId.HasValue)
            {
                if (!this.sessionRepository.IsSessionValid(sessionId.Value))
                {
                    return false;
                }

                var session = this.sessionRepository.GetSessionById(sessionId.Value);
                userId = session.UserId;
            }
            else
            {
                returnedSessionId = this.sessionRepository.CreateTempSession(userId).Id;
            }

            return true;
        }

        private void CopyUserStatsFromTempSession(Guid sessionId, Guid userId)
        {
            var session = this.sessionRepository.GetSessionById(sessionId);
            var stats = this.statisitcsRepository.GetUserStatistics(session.UserId);

            if (stats != null)
            {
                this.statisitcsRepository.MoveStatisticToNewUser(stats.Id, userId);
            }
        }

        private void RemoveGame(Guid id)
        {
            if (!games.ContainsGame(id))
            {
                logger.Warn("Game repository has no game with id: " + id);
                return;
            }

            games.RemoveGame(id);
            logger.Debug("Game removed from repository. Id: " + id);
        }

        private void InitService()
        {
            this.hashGenerator = new BCryptHashGenerator();
            Model.Initializer.Initialize();
            connectionPool.Start();
        }

        private void ResetQueue()
        {
            p1 = Guid.Empty;
            waitingForOpponentGameId = Guid.Empty;
        }

        private void FirstPlayerAvailabilityChecker()
        {
            while (!cts.IsCancellationRequested)
            {
                if (p1 != Guid.Empty &&
                    DateTime.Now - lastFirstPlayerPingDate > playerDisconnectionTimeout)
                {
                    logger.Info("First player is disconected and remover from waiting queue");
                    ResetQueue();
                    break;
                }

                Thread.Sleep(500);
            }
        }
    }
}
