﻿using System;
using System.Security.Principal;
using System.Web;
using System.Web.Security;

using TicTacToe.Service.DataContracts;
using TicTacToe.WebServiceHost.TicTacToeService;

namespace TicTacToe.WebServiceHost.Authentication
{
    public class ServiceAuthentication : IAuthentication
    {
        private string cookieName = "__AUTH__";
        private string sessionIdCookieName = "sessionId";

        private IPrincipal currentUser;

        public HttpContext HttpContext { get; set; }

        public IPrincipal CurrentUser
        {
            get
            {
                if (currentUser == null)
                {
                    try
                    {
                        HttpCookie authCookie = HttpContext.Request.Cookies.Get(cookieName);
                        if (authCookie != null && !string.IsNullOrEmpty(authCookie.Value))
                        {
                            var ticket = FormsAuthentication.Decrypt(authCookie.Value);
                            if (ticket != null)
                            {
                                currentUser = new UserProvider(ticket.Name, ticket.UserData);
                            }
                            else
                            {
                                currentUser = new UserProvider(null, null);
                            }
                        }
                        else
                        {
                            currentUser = new UserProvider(null, null);
                        }
                    }
                    catch (Exception)
                    {
                        // TODO: add logging.
                        currentUser = new UserProvider(null, null);
                    }
                }

                return currentUser;
            } 
        }

        public User Login(string email, string password)
        {
            using (var client = new TicTacToeServiceClient())
            {
                var result = client.Login(new UserLoginInfo { Login = email, Password = password });

                if (result.IsSuccess)
                {
                    CreateCookies(result.SessionId.ToString(), result.UserId.ToString());
                    return new User { SessionId = result.SessionId, UserId = result.UserId };
                }
            }

            return null;
        }

        public void Logout()
        {
            var httpCookies = HttpContext.Response.Cookies[cookieName];
            if (httpCookies != null)
            {
                httpCookies.Value = string.Empty;
            }

            var sessionCookies = HttpContext.Response.Cookies[sessionIdCookieName];
            if (sessionCookies != null)
            {
                sessionCookies.Value = string.Empty;
            }
        }

        private void CreateCookies(string sessionId, string userId)
        {
            var ticket = new FormsAuthenticationTicket(
                1,
                userId,
                DateTime.Now,
                DateTime.Now.Add(FormsAuthentication.Timeout),
                false,
                sessionId,
                FormsAuthentication.FormsCookiePath);

            var encTicket = FormsAuthentication.Encrypt(ticket);

            var authCookie = new HttpCookie(cookieName) { Value = encTicket, Expires = ticket.Expiration };

            HttpContext.Response.Cookies.Set(authCookie);

            var sessionIdCookie = new HttpCookie(sessionIdCookieName) { Value = sessionId, Expires = ticket.Expiration };
            HttpContext.Response.Cookies.Set(sessionIdCookie);
        }
    }
}